import pytest
import random

from genie_pkg.delimited_genie import generate, anonymise_columns
import csv
import math

type_choices = ['int', 'float', 'str', 'special_string']

def _fetch_column_index(colspecs, col_name):
    for i, c in enumerate(colspecs):
        data_type, *notused = c
        if col_name == data_type:
            return i    
    return

def test_csv():
    # Run all the tests in a loop to have better confidence of randomisation.
    for i in range(2, 5):
        ncols = random.randint(1, 10)
        default_specs = [(random.choice(type_choices),
                          random.randint(5, 15)) for i in range(ncols)]
        colspecs = default_specs + [('email', 15, 'mail.com'),
                                    ('date', '%d/%m/%Y', 3), 
                                    ('one_of', ['AU', 'NZ']),
                                    ('cc_mastercard',),
                                    ('cc_visacard', 13),
                                    ('geo_coord', (40.84, -73.87,),)]
        delimiter=random.choice([',', '|'])
        data = generate(colspecs, nrows=1, delimiter=delimiter)
        for d in data:
            decoded = d.decode()
            csv_data = list(csv.reader(decoded.splitlines(), delimiter=delimiter))[0]
            assert len(csv_data) == len(colspecs) + 1 # +1 is because geo_coord will produce 2 values
            visa_column_idx = _fetch_column_index(colspecs, 'cc_visacard')
            assert len(csv_data[visa_column_idx]) == 13
            country_idx = _fetch_column_index(colspecs, 'one_of')
            assert csv_data[country_idx] in ['AU', 'NZ']

def test_anonymise():
    input_encoding = 'windows-1252' 
    row = 'FReNG,£Ni,£iFthtR¥ubOswUPh,mQWJoypv,F¢MFcR,-37.814,144.963'.encode(input_encoding)

    anonymous_col_specs = [(1, 'float'), (4, 'int'), (5, 'geo_coord', (40.84, -73.87,), 1000)]
    anonymised = anonymise_columns(
        row, anonymous_col_specs, encoding=input_encoding)
    decoded = anonymised.decode(input_encoding)
    csv_data = list(csv.reader(decoded.splitlines()))[0]
    assert isinstance(int(csv_data[4]), int) == True
    assert isinstance(float(csv_data[1]), float) == True
    assert math.isclose(float(csv_data[5]), 40.84, abs_tol=0.010)


import pytest
import pandas as pd

from genie_pkg.dqc import QualityChecker
from hypothesis import given, settings, example
from hypothesis.strategies import composite, integers, lists, text, dates, just, one_of, sets
import string
from typing import List
import json
from datetime import date
import numpy as np

def test_validate_successful():
    df = pd.DataFrame([{'name': 'foo',
                        'dob': '1970-01-01'}])
    QualityChecker(df)


def test_validate_throws_exception():
    df = pd.DataFrame({'P': []})
    with pytest.raises(Exception) as _:
        QualityChecker(df)


def test_parse_error():
    df = pd.DataFrame([{'name': 'foo',
                        'd': '1970-01-01'}])
    check_spec = """
                apply checks {
                    foo
                }
                """

    with pytest.raises(Exception) as _:
        df.dqc.run(check_spec)


def _build_row_count(c: str, rhs: int):
    return f"row_count {c} {rhs}"


def _build_has_columns(columns: List[str], ignore_case=False):
    return "has_columns({}, ignore_case={})".format(json.dumps(columns), ignore_case)


def _build_column_is_date(column: str):
    return "is_date({})".format(column)


def _build_column_is_not_null(column: str):
    return "is_not_null({})".format(column)


def _build_column_has_one_of(column: str, values: List[str], ignore_case=False):
    return "has_one_of({}, {}, ignore_case={})".format(column, json.dumps(values), ignore_case)


def _build_column_has_positive_values(column: str):
    return "is_positive({})".format(column)


def _build_column_has_unique_values(column: str):
    return "is_unique({})".format(column)


def _build_quantile(column: str, q: float, c: str, rhs: float):
    return f"quantile({column}, {q}) {c} {rhs}"


def _build_percent_value_length(column: str, pass_percent_threshold: int, rhs: int, ignore_nulls=None):
    if ignore_nulls is not None:
        return f"percent_of_values_have_length({column}, pass_percent_threshold={pass_percent_threshold}, ignore_nulls={ignore_nulls}) == {rhs}"
    else:
        return f"percent_of_values_have_length({column}, pass_percent_threshold={pass_percent_threshold}) == {rhs}"

def _build_when_row(row_identifier, then_column, then_value):
    return f'when row_identified_by {json.dumps(row_identifier)} then {{"{then_column}" == "{then_value}"}}'


@composite
def generate_valid_checks(draw):
    row_count = draw(integers(min_value=20))
    columns = draw(lists(text(alphabet=string.ascii_letters + string.digits + '-_', min_size=1), min_size=1))
    column = draw(text(alphabet=string.ascii_letters, min_size=1))
    comparator = draw(one_of(just(">"), just("=="), just("<")))
    genders = ['female', 'male', 'other']
    bool = draw(one_of(just(False), just(True)))
    row_identifier = {"c1": "v1", "c2": "2"}
    return [
        _build_row_count(comparator, row_count),
        _build_has_columns(columns, ignore_case=bool),
        _build_column_is_date("dob"),
        _build_column_is_not_null(column),
        _build_column_has_one_of("gender", genders),
        _build_column_has_positive_values("age"),
        _build_column_has_unique_values("dob"),
        _build_quantile("field_A", 0.5, "==", 55.0),
        _build_quantile("field_A", 0.1, ">", 9.0),
        _build_quantile("field_A", 0.5, "<", 56.0),
        _build_percent_value_length("post_code", 50, 4, ignore_nulls=bool),
        _build_when_row(row_identifier, "c3", "v3"),
    ]


@given(generate_valid_checks(),
       lists(dates(min_value=date(1970, 1, 1), max_value=date(2100, 1, 1)).map(str), min_size=10, max_size=10),
       lists(integers(min_value=10), min_size=10, max_size=10),
       just([80, 24, 74, 30, 72, 69, 27, 12, 84, 41]),
       just(['female', 'male', 'other', 'female', 'male', 'other', 'female', 'male', 'other', 'other']),
       lists(integers(min_value=1000, max_value=9999), min_size=10, max_size=10), )
@settings(max_examples=101, deadline=500)
def test_hypothesis(checks, dobs, ages, q_arr, genders, post_codes):
    check_spec = """
                    apply checks {
                        %s
                    }
                    """ % ('\n'.join(checks),)

    df = pd.DataFrame({'age': ages,
                       'dob': dobs,
                       'field_A': q_arr,
                       'gender': genders,
                       'post_code': post_codes,
                       'c1': 'v1', 'c2': "2", 'c3': 'v3'})

    successes = list(filter(lambda x: x[2], df.dqc.run(check_spec)))
    # is_date, is_positive, quantile, has_one_of
    assert len(successes) > 0


@given(lists(integers(min_value=1000, max_value=9999), min_size=4, max_size=4))
@example(["3000", "0800", None, ""])
@settings(deadline=500)
def test_percent_value_length(postcodes):
    df = pd.DataFrame({'post_code': postcodes})
    check_spec = """
                apply checks {
                    percent_of_values_have_length(post_code, pass_percent_threshold=50) == 4
                }
                """
    successes = list(filter(lambda x: x[2], df.dqc.run(check_spec)))
    assert len(successes) > 0
    failures = list(filter(lambda x: not x[2], df.dqc.run(check_spec)))
    assert len(failures) >= 0


def test_validate_spec_returns_false():
    check_spec = """
                apply checks {
                }
                """
    err, res = QualityChecker.validate_spec(check_spec)
    assert err is not None
    assert not res


def test_validate_spec_returns_true():
    check_spec = """
                apply checks {
                    row_count > 0
                }
                """
    err, res = QualityChecker.validate_spec(check_spec)
    assert err is None
    assert res


@given(
    sets(text(alphabet=string.ascii_letters + '-_', min_size=3), min_size=5, max_size=5),
    lists(integers(min_value=1000, max_value=9999), min_size=5, max_size=5)
)
@settings(deadline=500)
def test_has_columns(columns, values):
    data = {}
    for c in columns:
        data[c] = values

    df = pd.DataFrame(data)
    check_spec_template = """
                    apply checks {
                        %s
                    }
                    """

    columns_lower_cased = [c.lower() for c in columns]
    check_spec_case_sensitive = check_spec_template % (_build_has_columns(columns_lower_cased, ignore_case=False),)
    result_case_sensitive = df.dqc.run(check_spec_case_sensitive)
    failures = list(filter(lambda x: not x[2], result_case_sensitive))
    assert len(failures) == 1
    successes = list(filter(lambda x: x[2], result_case_sensitive))
    assert len(successes) == 0
    check_spec_ignore_case = check_spec_template % (_build_has_columns(list(columns), ignore_case=True),)
    result_ignore_case = df.dqc.run(check_spec_ignore_case)
    failures = list(filter(lambda x: not x[2], result_ignore_case))
    assert len(failures) == 0
    successes = list(filter(lambda x: x[2], result_ignore_case))
    assert len(successes) == 1


def test_has_one_of():
    gender = ['female', 'male', 'other']
    data = {'Gender': ['Male', 'Female']}
    check_spec_template = """
                    apply checks {
                        %s
                    }
                    """

    df = pd.DataFrame(data)
    check_spec_case_sensitive = check_spec_template % (_build_column_has_one_of('gender', gender),)
    result_case_sensitive = df.dqc.run(check_spec_case_sensitive)
    failures = list(filter(lambda x: not x[2], result_case_sensitive))
    assert len(failures) == 1
    successes = list(filter(lambda x: x[2], result_case_sensitive))
    assert len(successes) == 0
    # previous test would have mutated the column name, hence recreate the df
    df = pd.DataFrame(data)
    check_spec_ignore_case = check_spec_template % (_build_column_has_one_of('gender', gender, ignore_case=True),)
    result_ignore_case = df.dqc.run(check_spec_ignore_case, ignore_column_case=True)
    failures = list(filter(lambda x: not x[2], result_ignore_case))
    assert len(failures) == 0
    successes = list(filter(lambda x: x[2], result_ignore_case))
    assert len(successes) == 1

def test_when_success():
    check_spec = """
                    apply checks {
                        when row_identified_by {
                            "c1": "v1",
                            "c2": "2"
                        } then {
                            "c3" == "v3",
                            "c4" == "v4"
                        }
                    }
                    """

    df = pd.DataFrame({"c1": "v1", "c2": "2", "c3": "v3", "c4": "v4"}, index=[0])
    result = df.dqc.run(check_spec)
    assert result[0][2]

def test_when_row_identifier_column_missing():
    check_spec = """
                    apply checks {
                        when row_identified_by {"c1": "v1", "c4": "2"} then {"c3" == "v3"}
                    }
                    """

    df = pd.DataFrame({"c1": "v1", "c2": "2", "c3": "v3"}, index=[0])
    with pytest.raises(Exception) as _:
        df.dqc.run(check_spec)

def test_when_target_column_missing():
    check_spec = """
                    apply checks {
                        when row_identified_by {"c1": "v1", "c2": "2"} then {"c3" == "v3"}
                    }
                    """

    df = pd.DataFrame({"c1": "v1", "c2": "2", "c4": "v3"}, index=[0])
    result = df.dqc.run(check_spec)
    assert not result[0][2]

def test_is_date_with_none_ignore_nulls():
    df = pd.DataFrame({'dob': ['1970-01-01', 'foo', None, np.nan]})
    check_spec = """
                apply checks {
                    is_date(dob, ignore_nulls=True)
                }
                """

    result = df.dqc.run(check_spec)
    assert not result[0][2]

def test_is_date_with_none_donot_ignore_nulls():
    df = pd.DataFrame({'dob': ['1970-01-01', 'foo', None, np.nan]})
    check_spec = """
                apply checks {
                    is_date(dob)
                }
                """

    result = df.dqc.run(check_spec)
    assert not result[0][2]

def test_is_date_success():
    df = pd.DataFrame({'dob': ['1970-01-01', 'foo', None, np.nan]})
    check_spec = """
                apply checks {
                    is_date(dob, pass_percent_threshold=50, ignore_nulls=True)
                    is_date(dob, pass_percent_threshold=25)
                }
                """

    result = df.dqc.run(check_spec)
    assert  result[0][2]

def test_is_date_division_by_zero():
    df = pd.DataFrame({'dob': [None, np.nan], 'bar': 'foo'})
    check_spec = """
                apply checks {
                    is_date(dob, ignore_nulls=True)
                }
                """

    result = df.dqc.run(check_spec)
    assert result[0][2]

def test_percent_values_have_length_param_combinations():
    df = pd.DataFrame([{'name': 'foo',
                        'd': '1970-01-01'}])
    check_spec = """
                apply checks {
                    percent_of_values_have_length(foo, pass_percent_threshold=50, ignore_nulls=True) == 2
                    percent_of_values_have_length(bar, ignore_nulls=True) == 2
                    percent_of_values_have_length(foobar, pass_percent_threshold=50) == 2
                    percent_of_values_have_length(name) == 3
                }
                """

    successes = list(filter(lambda x: x[2], df.dqc.run(check_spec)))
    assert len(successes) > 0



import pytest
import json

from genie_pkg.json_genie import generate, generate_with_custom_template_function
from jsonschema import validate
from jinja2 import Template
import random


def test_almighty():
    schema = {
        "type": "object",
        "properties": {
            "k1": {"type": "number"},
            "k2": {"type": "string"},
            "k3": {"type": "array", "items": {"type": "string"}},
            "k4": {"type": "number"},
            "k5": {"type": "boolean"},
            "k6": {"type": "number"},
            "k7": {"type": "string"},
            "k8": {
                "type": "array",
                "items": {
                    "anyOf": [
                        {
                            "type": "object",
                            "properties": {
                                "y1": {
                                    "type": "integer"
                                },
                                "y2": {
                                    "type": "integer"
                                }
                            },
                            "required": [
                                "y1"
                            ]
                        }
                    ]
                }
            },
            "k9": {"type": "array", "items": {"type": "integer"}},
            "k10": {"type": "string"},
            "k11": {"type": "string"},
            "k12": {"type": "string"},
            "k13": {"type": "string"},
            "k14": {"type": "string"},
            "k15": {
                "type": "object",
                "properties": {
                    "latitude":{"type": "number"},
                    "longitude":{"type": "number"}
                },
                "required": [
                    "latitude",
                    "longitude"
                ]
            },
            "k16": {"type": "string"},
            "k17": {"type": "string"},
            "k18": {"type": "string"}
        },
        "reauired": [
            "k1", "k2", "k3", "k18"
        ]
    }

    template = """
        {
            "k1": {{random_integer(1, 1000)}},
            "k2": "{{random_string_with_special_chars(5)}}",
            "k3": {{random_string_list(2, 8)}},
            "k4": {{random_float(1, 100, 2)}},
            "k5": {{random_bool()}},
            "k6": {{now_epoch()}},
            "k7": "{{random_string(5)}}",
            "k8": [
                {
                    "y1": {{random_integer(20, 40)}}
                },
                {
                    "y1": {{random_integer(20, 60)}}
                }
            ],
            "k9": {{random_integer_list(2, 60, 100)}},
            "k11": "{{random_choice_of(['apple', 'mango'])}}",
            "k10": "{{guid()}}",
            "k12": "{{date_with_format('%d/%m/%Y', -10)}}",
            "k13": "{{random_email_id(20, 'gmail.com')}}",
            "k14": "{{random_ipv4()}}",
            "k15": {{random_geo(accuracy=4)}},
            "k16": "{{random_mastercard_number()}}",
            "k17": "{{random_visacard_number()}}",
            "k18": "{{random_text()}}"
        }
    """
    d = json.loads(generate(template))
    validate(instance=d, schema=schema)


fruit_choices = ['mango', 'apple', 'durian', 'jackfruit']


def favourite_fruit():
    return random.choice(fruit_choices)


def test_generate_with_custom_template_function():
    template = """
        {
            "k1": {{random_integer(1, 1000)}},
            "k2": "{{favourite_fruit()}}"
        }
    """

    schema = {
        "type": "object",
        "properties": {
            "k1": {"type": "number"},
            "k2": {"type": "string", "enum": fruit_choices}
        }
    }

    t = Template(template)
    t.globals['favourite_fruit'] = favourite_fruit

    d = json.loads(generate_with_custom_template_function(t))
    validate(instance=d, schema=schema)
